@@ -36,9 +36,18 @@ import os, pwd, sys, time, MySQLdb, signal
 os.chdir("/")
 
 foreground = 0
-DEBUG = 0
 
-expire_interval = 300
+# DEBUG: very verbose logs
+DEBUG = False
+
+# DRYRUN: do not terminate any sessions, and check often 
+DRYRUN = False
+
+if DRYRUN:
+  expire_interval = 30
+else:
+  expire_interval = 300
+  
 logfile_name = "/var/log/expire-sessions/expire-sessions.log"
 logfile = None
 pidfile = "/var/run/expire-sessions.pid"
@@ -74,6 +83,7 @@ mysql_db = ""
 mysql_connect_attempts = 120
 
 ANONYMOUS_SESSIONS = False
+HAS_END_JOBLOG = False
 
 #=============================================================================
 # Load the configuration and override the variables above.
@@ -272,14 +282,25 @@ if not foreground:
 
   log("Server is ready.")
 
-# Delete old views without heartbeat
 db = db_connect()
 c = db.cursor()
 
+# Check whether calculation using joblog "end" timestamp can be used
+try:
+  c.execute("""select end from joblog limit 1""" )
+except:
+  HAS_END_JOBLOG = False
+  log("joblog does not have end timestamp, using less accurate calculation with walltime")
+else:
+  HAS_END_JOBLOG = True
+  log("joblog has end timestamp")
+  
 while 1:
   # Recheck, at the longest, every "expire_interval" seconds.
   sleeptime = expire_interval
   # delete any orphaned views, which shouldn't exist unless there was a restart or crash
+  if DEBUG:
+    log("deleting dead views")
   mysql(c, """DELETE FROM view WHERE heartbeat < DATE_ADD(NOW(), INTERVAL -1 DAY)""")
 
   # Anonymous sessions
@@ -299,9 +320,11 @@ while 1:
   # Find sessions that have been accessed more than "timeout" seconds ago
   # and that have no views, and that could expire during our next sleep (expire_interval)
   # skew: keep track of sleep time between each expired session
-  skew = 0
+  #     AND zone_id > 0 :  not all hubs have zones, some will have a zone_id =0
+  skew = 0 
   arr = mysql(c, """
-    SELECT session.sessnum, session.username, appname,
+    SELECT session.sessnum, session.username, appname, timeout, 
+      NOW() as curtime,
       timeout - TIMESTAMPDIFF(SECOND, accesstime, NOW()) AS remaining
     FROM session LEFT JOIN view ON session.sessnum = view.sessnum
     WHERE viewid IS NULL
@@ -309,57 +332,101 @@ while 1:
     AND timeout - TIMESTAMPDIFF(SECOND, accesstime, NOW()) < %d
     ORDER BY remaining""" % (expire_interval))
 
-  # Check if those sessions have jobs running
+  # session is a candidate for termination if remaining  <=0
+  # other sessions are used to calculate sleep time for this daemon
+  #
+  # Check if candidates for termination have jobs running
   # there can be many jobs per session.  Expire session only if *all* jobs have had no heartbeat.
   # Complication:  submit --local jobs that run as long as the session is open
   # "submit --local" will have no heartbeat, whereas all other submit jobs will
   # To handle this, expire session only if all job heartbeats are old.
-  for sessnum, username, appname, session_remaining in arr:
+  # we keep track of curtime during which the query was run for later calculations, because NOW() would keep increasing
+  # whereas we want a snapshot in time of events to calculate timeouts
+  for sessnum, username, appname, timeout, curtime, session_remaining in arr:
     # Find the greatest remaining time for this session to stay alive
     max_remaining = int(session_remaining)
+    if DEBUG:
+      log("found session %d without views, with %d seconds remaining out of %d timeout" % (sessnum, max_remaining, int(timeout)))
     # short circuit other checks if this session will stay alive longer than interval to next check
     if max_remaining > sleeptime:
       continue
-    # heartbeat is considered "stale" after 24 hours
-    # remaining = heartbeat - NOW() + 86400
+    #
+    # heartbeat is considered "stale" after timeout seconds
+    # remaining = heartbeat + timeout - NOW() = timeout - (NOW() - heartbeat)
     # In submit version >= 2.7.x job.active=2 jobs should be ignored when considering the heartbeat update
-    job_arr = mysql(c, """SELECT 86400-TIMESTAMPDIFF(SECOND, heartbeat, NOW()) AS remaining
+    job_arr = mysql(c, """SELECT %d-TIMESTAMPDIFF(SECOND, heartbeat, "%s") AS remaining
       FROM job WHERE sessnum = "%s" AND active !=2
-      AND 86400-TIMESTAMPDIFF(SECOND, heartbeat, NOW()) > 0""" % sessnum)
+      AND %d-TIMESTAMPDIFF(SECOND, heartbeat, "%s") > 0""" % (timeout, curtime, sessnum, timeout, curtime))
     if job_arr:
       # Find the greatest reason this session should stay alive
       for row in job_arr:
         remaining = row[0]
+        if DEBUG:
+          log("found job with %d seconds remaining out of %d timeout" % (int(remaining), int(timeout)))
         if int(remaining)  > max_remaining:
           max_remaining = int(remaining)
+        if max_remaining > sleeptime:
+          # don't look at any more jobs, this session's survival is assured
+          break
       if max_remaining > sleeptime:
+	# skip to next session
         continue
 
-    # How long should someone have to check the results of a simulation?
-    # people have 24 hours to check results after the end of the last job.
-    # If a job started at 9AM took 12 hours to run then a user would have until 9PM
-    # the next day to check results.  It's always possible to check the results before
-    # the session ends.
-    # if any joblog entry is for a job that ended less than a day ago, pass
-    # remaining = start + walltime - NOW() + 86400
-    # walltime > TIME_TO_SEC(start) is because it's possible for walltime to be calculated from epoch
+    # How long should someone have to check the results of a simulation?  
+    # people have <timeout> seconds to check results after the end of the last job.  
+    # if any joblog entry is for a job that ended less than <timeout>, pass
+    # remaining time until timeout 
     # job > 0 because the middleware makes entries with job=0, we don't want those, not real jobs
-    job_arr = mysql(c, """SELECT 86400 + walltime - TIMESTAMPDIFF(SECOND, start, NOW()) AS remaining
+    #
+    # calculation using "end" timestamp instead of walltime
+    # alter table joblog add column end timestamp  DEFAULT CURRENT_TIMESTAMP;
+    if HAS_END_JOBLOG:
+      job_arr = mysql(c, """SELECT %d - TIMESTAMPDIFF(SECOND, end, "%s") AS remaining
+        FROM joblog WHERE sessnum = "%s"
+        AND job > 0
+        AND end > '0000-00-00 00:00:00'
+        AND %d - TIMESTAMPDIFF(SECOND, end, "%s") > 0""" % (timeout, curtime, sessnum, timeout, curtime))
+      if job_arr:
+        for row in job_arr:
+          remaining = row[0]
+          if int(remaining)  > max_remaining:
+            max_remaining = int(remaining)
+          if DEBUG:
+            log("%d of %s has a recently ended job with %d seconds left of grace to check results" % (sessnum, username, row[0]))
+
+    #
+    # calculation using "walltime", which may be wildly inaccurate, 
+    # which is why the "end" column was added with an automatic timestamp
+    # doing this calculation can only help prevent expiring a session too early so it does no harm
+    # walltime here is supposed to be in seconds;  the condition "walltime < UNIX_TIMESTAMP(start)" prevents cases
+    # where the walltime is instead a timestamp.
+    job_arr = mysql(c, """SELECT %d + walltime - TIMESTAMPDIFF(SECOND, start, "%s") AS remaining
       FROM joblog WHERE sessnum = "%s"
       AND walltime < UNIX_TIMESTAMP(start)
       AND job > 0
-      AND 86400 + walltime - TIMESTAMPDIFF(SECOND, start, NOW()) > 0""" % sessnum)
+      AND end = '0000-00-00 00:00:00'
+      AND %d + walltime - TIMESTAMPDIFF(SECOND, start, "%s") > 0""" % (timeout, curtime, sessnum, timeout, curtime))
     if job_arr:
       for row in job_arr:
         remaining = row[0]
         if int(remaining)  > max_remaining:
           max_remaining = int(remaining)
-      # log("%d of %s has a recently ended job %s" % (sessnum, username, job_arr))
+        if DEBUG:
+          log("%d of %s has a recently ended job with %d seconds left of grace to check results" % (sessnum, username, row[0]))
 
-    if DEBUG:
-      log("%d of %s to expire in %d seconds" % (sessnum, username, max_remaining))
-    if max_remaining < 1 + skew:
-      terminate(sessnum, username, appname)
+    if max_remaining > sleeptime:
+      if DEBUG:
+        log("%d of %s to expire in %d seconds" % (sessnum, username, max_remaining))
+      # skip to next session
+      continue
+    if max_remaining < 0:
+      # this session should be terminated
+      # do not adjust this decision with skew because 
+      # another event could have happened during <skew> seconds, that would preserve session
+      if DRYRUN:
+        log("would be terminating %d (%s,%s)" % (sessnum, username, appname))
+      else:
+        terminate(sessnum, username, appname)
       time.sleep(1) # Do not kill things off too quickly.
       skew += 1
     elif max_remaining < sleeptime:
@@ -368,6 +435,8 @@ while 1:
   sleeptime -= skew
   if (sleeptime > 0):
     db.close()
+    if DEBUG:
+      log("sleeping %d" % sleeptime)
     time.sleep(sleeptime)
     db = db_connect()
     c = db.cursor()
